@@ -12,7 +12,7 @@
 
    ---------------------------------------------------------------------------
 
-   Copyright (C) 2003,2004,2005  Volker C. Behr
+   Copyright (C) 2003,2004,2005,2006  Volker C. Behr
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
@@ -34,17 +34,43 @@
    has to be preserved and all modifications should be clearly 
    indicated.
    In case you want to include this code into your own programs 
-   I would appreaciated some feedback via email.
+   I would appreciate your feedback via email.
 
   
    HISTORY: 
+   2006-04-14 : 2.2.0      - new option to remove prefixes from usernames
+   2006-04-06 : 2.1.1      - minor bugfix to read umasks properly
+   2006-03-25 : 2.1.0	   - new option to select where to look for title,
+                             some code changes and optimizations
+   2006-02-26 : 2.0.5      - several code optimizations
+   2006-01-29 : 2.0.4      - beginning of (E)PS-code now determined by "%!",
+                             optimized system call to GS in cups-pdf.conf,
+              		     some minor code changes,
+			     debug information shows version number,
+      			     updated PostScript PPD file supporting ledger paper
+   2006-01-07 : 2.0.3	   - improved checking for root privileges (incl. EUID)
+   2006-01-04 : 2.0.2	   - fixed possible testing of an uninitialized variable
+   2006-01-03 : 2.0.1	   - fixed hanging jobs when logging is disabled
+   2005-12-23 : 2.0.0      - creation of non-existent parent directories,
+                             improved configurability of output directories,
+			     option CPHOMESUB is now obsolete,
+                             PPD file now matches Adobe specifications
+   2005-10-23 : 2.0beta2   - added support for post-processing script,
+                             fixed a bug when parsing cups-pdf.conf
+   2005-09-18 : 2.0beta1   - added support for runtime configuration,
+                             introduced cups-pdf.conf,
+                             made AnonDirName an abolute path
+   2005-09-08 : 1.7.4      - added some options to cups-pdf.h for MacOSX
+   2005-07-23 : 1.7.3      - fixed GhostScript failure with error 11,
+                             fixed segfault when called without root privileges
+   2005-07-07 : 1.7.2	   - adjustment of file mode moved to non-privileged child
    2005-06-16 : 1.7.1	   - PDFs are protected during creation by file mode 000
    2005-03-07 : 1.7.0a
    2005-03-01 : 1.7.0      - call of GhostScript from a non-privileged child process,
                              no more additional programs in the call of GhostScript,
                              several code optimizations
    2005-02-09 : 1.6.6      - updated PostScript PPD file supporting more DIN formats
-   2005-01-04 : 1.6.5      - cups-pdf is running with gid of group defined in CPGRP
+   2005-01-04 : 1.6.5      - cups-pdf is running with gid of group defined in conf.grp
    2004-10-11 : 1.6.4      - improved configurability for different versions of 'su'
    2004-10-07 : 1.6.3      - corrected erroneous call of GhostScript
    2004-10-05 : 1.6.2      - added necessary quotes in the call of GhostScript
@@ -64,7 +90,7 @@
    2004-02-09 : 1.4.1	   - logfile is created with mode 644 if it does not exist
    2004-01-27 : 1.4.0      - new option to set permissions for PDF output,
                              alternative options for improved image conversion
-                             (moved CPGSCALL to cups-pdf.h)
+                             (moved conf.gscall to cups-pdf.h)
    2004-01-06 : 1.3.2	   - exit status of ghostscript gets logged in debug mode
    2004-01-02 : 1.3.1a
    2003-11-29 : 1.3.1	   - improved handling of domain prefix
@@ -88,13 +114,6 @@
    2003-02-14 : 0.2    - script re-located to /usr/lib/cups/filter/cups-pdfgen
    2003-02-12 : 0.1    - first beta release			                */
 
-#define CPERROR 	1
-#define CPSTATUS 	2
-#define CPDEBUG 	4
-
-#define BUFSIZE	1024
-#define TBUFSIZE "1024"
-
 #include <time.h>
 #include <errno.h>
 #include <stdio.h>
@@ -114,13 +133,13 @@
 
 extern int errno;
 
-FILE *logfp;
+FILE *logfp=NULL;
 
 void log_event(char type, char message[48], char *detail) {
   time_t secs;
   char ctype[8], *timestring;
 
-  if (strlen(CPLOG) && (type & CPLOGTYPE)) { 
+  if (strlen(conf.log) && (type & conf.logtype)) { 
     time(&secs);
     timestring=ctime(&secs);
     timestring[strlen(timestring)-1]='\0';
@@ -135,7 +154,7 @@ void log_event(char type, char message[48], char *detail) {
       while (detail[strlen(detail)-1] == '\n') 
         detail[strlen(detail)-1]='\0';
       fprintf(logfp,"%s  [%s] %s (%s)\n", timestring, ctype, message, detail);
-      if ((CPLOGTYPE & CPDEBUG) && (type & CPERROR)) 
+      if ((conf.logtype & CPDEBUG) && (type & CPERROR)) 
         fprintf(logfp,"%s  [DEBUG] ERRNO: %d\n", timestring, errno);
     }
     else
@@ -143,74 +162,182 @@ void log_event(char type, char message[48], char *detail) {
   }
   return;
 }
+
+void read_config() {
+  FILE *fp;
+  cp_string buffer, key, value;
+
+  fp=fopen(CPCONFIG,"r");
+  if (fp==NULL)
+    return;
+
+  while (fgets(buffer, BUFSIZE, fp) != NULL) {
+    key[0]='\0';
+    value[0]='\0';
+    if (sscanf(buffer,"%s %[^\n]",key,value)) {
+      if (!strlen(key) || !strncmp(key,"#",1))
+        continue;
+      if (!_assign_value(key, value))
+        fputs("Error in configuration file!\n", stderr);
+    }
+  }
+
+  fclose(fp);
+  return;
+}
+
+int create_dir(char *dirname, int nolog) {
+  struct stat fstatus;
+  char buffer[BUFSIZE],*delim;
+  int i;
+
+  while ((i=strlen(dirname))>1 && dirname[i-1]=='/')
+    dirname[i-1]='\0';
+  if (stat(dirname, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
+    memcpy(buffer,dirname,BUFSIZE-1);
+    delim=strrchr(buffer,'/');
+    if (delim!=buffer)
+      delim[0]='\0';
+    else
+      delim[1]='\0';
+    if (create_dir(buffer,nolog)!=0)
+      return 1;
+    stat(buffer, &fstatus);
+    if (mkdir(dirname,fstatus.st_mode)!=0) {
+      if (!nolog)
+        log_event(CPERROR, "failed to create directory", dirname);
+      return 1;
+    }
+    else
+      if (!nolog)
+        log_event(CPSTATUS, "directory created", dirname);
+    if (chown(dirname,fstatus.st_uid,fstatus.st_gid)!=0)
+      if (!nolog)
+        log_event(CPDEBUG, "failed to set mode on directory (non fatal)", dirname);
+  }
+  return 0;
+}
     
 int init() {
   struct stat fstatus;
   struct group *group;
+  cp_string logfile;
+
+  _set_defaults();
+  if (strlen(CPCONFIG))
+    if (!stat(CPCONFIG, &fstatus) && 
+        (S_ISREG(fstatus.st_mode) || S_ISLNK(fstatus.st_mode)))
+      read_config();
 
   umask(0077);
 
-  group=getgrnam(CPGRP);
+  group=getgrnam(conf.grp);
   if (group)
     setgid(group->gr_gid);
 
-  if (strlen(CPLOG)) {
-    if (stat(CPLOG, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
-      if (mkdir(CPLOG, 0700))
+  if (strlen(conf.log)) {
+    if (stat(conf.log, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
+      if (create_dir(conf.log, 1)) 
         return 1;
-      log_event(CPSTATUS, "log directory created", CPLOG);
+      if (chmod(conf.log, 0700))
+        return 1;
+      log_event(CPSTATUS, "log directory created", conf.log);
     }
-    logfp=fopen(CPLOG"/cups-pdf_log", "a");
+    snprintf(logfile,BUFSIZE,"%s%s",conf.log,"/cups-pdf_log");
+    logfp=fopen(logfile, "a");
   }
 
   if (!group) {
-    log_event(CPERROR, "CPGRP not found", CPGRP);
+    log_event(CPERROR, "Grp not found", conf.grp);
     return 1;
   }
   else 
-    log_event(CPDEBUG, "switching to new gid", CPGRP);
+    log_event(CPDEBUG, "switching to new gid", conf.grp);
 
   umask(0022);
-  if (!strncmp(CPOUT, "$HOME", 5)) 
-    log_event(CPDEBUG, "output redirected to home directory", "$HOME");
-  else if (stat(CPOUT, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
-    if (mkdir(CPOUT, 0755)) {
-      log_event(CPERROR, "failed to create output directory", CPOUT);
-      return 1;
-    }
-    log_event(CPSTATUS, "output directory created", CPOUT);
-  }
 
-  if (stat(CPSPOOL, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
-    if (mkdir(CPSPOOL, 0755)) {
-      log_event(CPERROR, "failed to create spool directory", CPSPOOL);
+  if (stat(conf.spool, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
+    if (create_dir(conf.spool, 0)) {
+      log_event(CPERROR, "failed to create spool directory", conf.spool);
       return 1;
     }
-    log_event(CPSTATUS, "spool directory created", CPSPOOL);
+    if (chmod(conf.spool, 0755)) {
+      log_event(CPERROR, "failed to set mode on spool directory", conf.spool);
+      return 1;
+    }
+    if (chown(conf.spool, -1, group->gr_gid)) 
+      log_event(CPERROR, "failed to set group id on spool directory (non fatal)", conf.grp);
+    log_event(CPSTATUS, "spool directory created", conf.spool);
   }
 
   umask(0077);
   return 0;
 }
 
+char *preparedirname(struct passwd *passwd, char *uname) {
+  int size;
+  char bufin[BUFSIZE], bufout[BUFSIZE], *needle, *cptr;
+
+  needle=strstr(uname, conf.removeprefix);
+  if (strlen(uname)>(unsigned int)(size=strlen(conf.removeprefix)))
+    uname=uname+size;
+
+  strncpy(bufin, conf.out, BUFSIZE);
+  do {
+    needle=strstr(bufin, "${HOME}");
+    if (needle == NULL)
+      break;
+    needle[0]='\0';
+    cptr=needle+7;
+    snprintf(bufout, BUFSIZE, "%s%s%s", bufin, passwd->pw_dir, cptr);
+    strncpy(bufin, bufout, BUFSIZE);
+  } while (needle != NULL);
+  do {
+    needle=strstr(bufin, "${USER}");
+    if (needle == NULL)
+      break;
+    needle[0]='\0';
+    cptr=needle+7;
+    if (!conf.dirprefix)
+      snprintf(bufout, BUFSIZE, "%s%s%s", bufin, uname, cptr);
+    else
+      snprintf(bufout, BUFSIZE, "%s%s%s", bufin, passwd->pw_name, cptr);
+    strncpy(bufin, bufout, BUFSIZE);
+  } while (needle != NULL);
+  size=strlen(bufin)+1;
+  cptr=calloc(size, sizeof(char));
+  if (cptr == NULL)
+    return NULL;
+  snprintf(cptr,size,"%s",bufin);
+  return cptr;
+}
+
 int prepareuser(struct passwd *passwd, char *dirname) {
   struct stat fstatus;
 
   umask(0000);
   if (stat(dirname, &fstatus) || !S_ISDIR(fstatus.st_mode)) {
-    if (!strcmp(passwd->pw_name, CPANONUSER)) {
-      if (mkdir(dirname, (0777&~CPANONUMASK))) {
+    if (!strcmp(passwd->pw_name, conf.anonuser)) {
+      if (create_dir(dirname, 0)) {
         log_event(CPERROR, "failed to create anonymous output directory", dirname);
         return 1;
       }
-      log_event(CPSTATUS, "anonymous output directory created", dirname);
+      if (chmod(dirname, (0777&~conf.anonumask))) {
+        log_event(CPERROR, "failed to set mode on anonymous output directory", dirname);
+        return 1;
+      }
+      log_event(CPDEBUG, "anonymous output directory created", dirname);
     }
     else {
-      if (mkdir(dirname, (0777&~CPUSERUMASK))) {
+      if (create_dir(dirname, 0)) {
         log_event(CPERROR, "failed to create user output directory", dirname);
         return 1;
       }
-      log_event(CPSTATUS, "user output directory created", dirname);
+      if (chmod(dirname, (0777&~conf.userumask))) {
+        log_event(CPERROR, "failed to set mode on user output directory", dirname);
+        return 1;
+      }
+      log_event(CPDEBUG, "user output directory created", dirname);
     }
     if (chown(dirname, passwd->pw_uid, passwd->pw_gid)) {
       log_event(CPERROR, "failed to set owner for output directory", passwd->pw_name);
@@ -234,10 +361,31 @@ void replace_string(char *string) {
     string[i]='_';
   return;
 }
+
+int preparetitle(char *title) {
+  char *cut;
   
-int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, int job, struct passwd *passwd) {
-  char buffer[BUFSIZE], *cut;
-  int rec_depth;
+  cut=strrchr(title, '/');
+  if (cut != NULL) {
+    log_event(CPDEBUG, "removing slashes from full title", title);
+    memmove(title, cut+1, strlen(cut+1)+1);
+  }  
+  if (title != NULL) 
+    while (strlen(title) && ((title[strlen(title)-1] == '\n') || (title[strlen(title)-1] == '\r'))) 
+      title[strlen(title)-1]='\0';
+  cut=strrchr(title, '.');
+  if ((cut != NULL) && ((short)strlen(cut) <= conf.cut+1) && (cut != title)) {
+    log_event(CPDEBUG, "removing file name extension", cut);
+    cut[0]='\0';
+  }
+  replace_string(title);
+  return strcmp(title, "");
+}
+
+int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, char *cmdtitle, 
+                     int job, struct passwd *passwd) {
+  cp_string buffer;
+  int rec_depth,is_title=0;
   FILE *fpdest;
 
   if (fpsrc == NULL) {
@@ -259,7 +407,7 @@ int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, int job, struct
   log_event(CPDEBUG, "owner set for spoolfile", spoolfile);
   rec_depth=0;
   while (fgets(buffer, BUFSIZE, fpsrc) != NULL) {
-    if (!strncmp(buffer, "%!PS", 4)) {
+    if (!strncmp(buffer, "%!", 2)) {
       log_event(CPDEBUG, "found beginning of postscript code", buffer);
       break;
     }
@@ -268,9 +416,12 @@ int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, int job, struct
   fputs(buffer, fpdest);
   while (fgets(buffer, BUFSIZE, fpsrc) != NULL) {
     fputs(buffer, fpdest);
-    if (!rec_depth)
-      sscanf(buffer, "%%%%Title: %"TBUFSIZE"c", title);
-    if (!strncmp(buffer, "%!PS", 4)) {
+    if (!is_title && !rec_depth)
+      if (sscanf(buffer, "%%%%Title: %"TBUFSIZE"c", title)==1) {
+        log_event(CPDEBUG, "found title in ps code", title);
+        is_title=1;
+      }
+    if (!strncmp(buffer, "%!", 2)) {
       log_event(CPDEBUG, "found embedded (e)ps code", buffer);
       rec_depth++;
     }
@@ -288,26 +439,36 @@ int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, int job, struct
   fclose(fpdest);
   fclose(fpsrc);
   log_event(CPDEBUG, "all data written to spoolfile", spoolfile);
-  cut=strrchr(title, '/');
-  if (cut != NULL) {
-    log_event(CPDEBUG, "removing slashes from full title", title);
-    memmove(title, cut+1, strlen(cut+1)+1);
-  }  
-  if (title != NULL) 
-    while (strlen(title) && ((title[strlen(title)-1] == '\n') || (title[strlen(title)-1] == '\r'))) 
-      title[strlen(title)-1]='\0';
-  cut=strrchr(title, '.');
-  if ((cut != NULL) && (strlen(cut) <= CPCUT+1) && (cut != title)) {
-    log_event(CPDEBUG, "removing file name extension", cut);
-    cut[0]='\0';
+
+  if (!strcmp(cmdtitle, "(stdin)")) 
+    cmdtitle="";
+  if (conf.titlepref) {
+    log_event(CPDEBUG, "trying to use commandline title", cmdtitle);
+    if (!preparetitle(cmdtitle)) {
+      log_event(CPDEBUG, "empty commandline title, using PS title", title);
+      if (!preparetitle(title))
+        log_event(CPDEBUG, "empty PS title", NULL); 
+    }
+    else
+      snprintf(title, BUFSIZE, "%s", cmdtitle);
   }
-  replace_string(title);
+  else {
+    log_event(CPDEBUG, "trying to use PS title", title);
+    if (!preparetitle(title)) {
+      log_event(CPDEBUG, "empty PS title, using commandline title", cmdtitle);
+      if (!preparetitle(cmdtitle))
+        log_event(CPDEBUG, "empty commandline title", NULL);
+      else
+        snprintf(title, BUFSIZE, "%s", cmdtitle);
+    }
+  }          
+
   if (!strcmp(title, "")) { 
     snprintf(title, BUFSIZE, "job_%i-untitled_document", job);
     log_event(CPDEBUG, "no title found - using default value", title);
   }
   else {
-    if (CPLABEL) {
+    if (conf.label) {
       strcpy(buffer, title);
       snprintf(title, BUFSIZE, "job_%i-%s", job, buffer);
     }
@@ -317,14 +478,21 @@ int preparespoolfile(FILE *fpsrc, char *spoolfile, char *title, int job, struct
 }
 
 int main(int argc, char *argv[]) {
-  char *user, *dirname, *spoolfile, *outfile, *gscall, title[BUFSIZE]="";
+  char *user, *dirname, *spoolfile, *outfile, *gscall, *ppcall;
+  cp_string title="";
   int size;
   mode_t mode;
   struct passwd *passwd;
   pid_t pid;
 
-  if (init()) return -1;
-  log_event(CPDEBUG, "initialization finished", NULL);
+  if (setuid(0)) {
+    fputs("cups-pdf cannot be called without root privileges!\n", stderr);
+    return 0;
+  }
+
+  if (init()) 
+    return -1;
+  log_event(CPDEBUG, "initialization finished", CPVERSION);
 
   if (argc==1) {
     printf("file cups-pdf:/ \"PDF Printer\" \"Virtual Printer\"\n"); 
@@ -333,125 +501,114 @@ int main(int argc, char *argv[]) {
   }
   if (argc<6 || argc>7) {
     fputs("Usage: cups-pdf job-id user title copies options [file]\n", stderr);
-    log_event(CPSTATUS, "call contained illegal number of arguments", NULL);
-  return 0;
+    log_event(CPERROR, "call contained illegal number of arguments", NULL);
+    return 0;
   }
-  size=strlen(CPUSERPREFIX)+strlen(argv[2])+1;
+
+  size=strlen(conf.userprefix)+strlen(argv[2])+1;
   user=calloc(size, sizeof(char));
   if (user == NULL) {
     fputs("CUPS-PDF: failed to allocate memory\n", stderr);
     return -2;
   }  
-  snprintf(user, size, "%s%s", CPUSERPREFIX, argv[2]);
+  snprintf(user, size, "%s%s", conf.userprefix, argv[2]);
   passwd=getpwnam(user);
-  if (passwd == NULL && CPLOWERCASE) {
+  if (passwd == NULL && conf.lowercase) {
     log_event(CPDEBUG, "unknown user", user);
     for (size=0;size<(int) strlen(argv[2]);size++) 
       argv[2][size]=tolower(argv[2][size]);
     log_event(CPDEBUG, "trying lower case user name", argv[2]);
-    size=strlen(CPUSERPREFIX)+strlen(argv[2])+1;
-    snprintf(user, size, "%s%s", CPUSERPREFIX, argv[2]);
+    size=strlen(conf.userprefix)+strlen(argv[2])+1;
+    snprintf(user, size, "%s%s", conf.userprefix, argv[2]);
     passwd=getpwnam(user);
   }  
   if (passwd == NULL) {
-    if (strlen(CPANONUSER)) {
-      passwd=getpwnam(CPANONUSER);
+    if (strlen(conf.anonuser)) {
+      passwd=getpwnam(conf.anonuser);
       if (passwd == NULL) {
-        log_event(CPERROR, "username for anonymous access unknown", CPANONUSER);
+        log_event(CPERROR, "username for anonymous access unknown", conf.anonuser);
         free(user);
-        fclose(logfp);
+        if (logfp!=NULL)
+          fclose(logfp);
         return -2;
       }
       log_event(CPDEBUG, "unknown user", user);
-      if (!strncmp(CPOUT, "$HOME", 5)) {
-        size=strlen(passwd->pw_dir)+strlen(CPHOMESUB)+4;
-        dirname=calloc(size, sizeof(char));
-        if (dirname == NULL) {
-          fputs("CUPS-PDF: failed to allocate memory\n", stderr);
-          free(user);
+      size=strlen(conf.anondirname)+4;
+      dirname=calloc(size, sizeof(char));
+      if (dirname == NULL) {
+        fputs("CUPS-PDF: failed to allocate memory\n", stderr);
+        free(user);
+        if (logfp!=NULL)
           fclose(logfp);
-          return -2;
-        }  
-        snprintf(dirname, size, "%s/%s", passwd->pw_dir, CPHOMESUB);
-      }
-      else {
-        size=strlen(CPOUT)+strlen(CPANONDIRNAME)+4;
-        dirname=calloc(size, sizeof(char));
-        if (dirname == NULL) {
-          fputs("CUPS-PDF: failed to allocate memory\n", stderr);
-          free(user);
-          fclose(logfp);
-          return -2;
-        }  
-        snprintf(dirname, size, "%s/%s", CPOUT, CPANONDIRNAME);
-      }
+        return -2;
+      }  
+      snprintf(dirname, size, "%s", conf.anondirname);
+      while (strlen(dirname) && ((dirname[strlen(dirname)-1] == '\n') ||
+             (dirname[strlen(dirname)-1] == '\r')))
+        dirname[strlen(dirname)-1]='\0';
+      log_event(CPDEBUG, "output directory name generated", dirname);
     }
     else {
       log_event(CPSTATUS, "anonymous access denied", user);
       free(user);
-      fclose(logfp);
+      if (logfp!=NULL)
+        fclose(logfp);
       return 0;
     }
-    mode=(0666&~CPANONUMASK);
+    mode=(0666&~conf.anonumask);
   } 
   else {
     log_event(CPDEBUG, "user identified", passwd->pw_name);
-    if (!strncmp(CPOUT, "$HOME", 5)) {
-      size=strlen(passwd->pw_dir)+strlen(CPHOMESUB)+4;
-      dirname=calloc(size, sizeof(char));
-      if (dirname == NULL) {
-        fputs("CUPS-PDF: failed to allocate memory\n", stderr);
-        free(user);
+    if ((dirname=preparedirname(passwd, argv[2])) == NULL) {
+      fputs("CUPS-PDF: failed to allocate memory\n", stderr);
+      free(user);
+      if (logfp!=NULL)
         fclose(logfp);
-        return -2;
-      }  
-      snprintf(dirname, size, "%s/%s", passwd->pw_dir, CPHOMESUB);
-    }
-    else {
-      size=strlen(CPOUT)+strlen(passwd->pw_name)+4;
-      dirname=calloc(size, sizeof(char));
-      if (dirname == NULL) {
-        fputs("CUPS-PDF: failed to allocate memory\n", stderr);
-        free(user);
-        fclose(logfp);
-        return -2;
-      }  
-      if (!CPDIRPREFIX)
-        snprintf(dirname, size, "%s/%s", CPOUT, argv[2]);
-      else
-        snprintf(dirname, size, "%s/%s", CPOUT, passwd->pw_name);
-    }
-    mode=(0666&~CPUSERUMASK);
+      return -2;
+    }  
+    while (strlen(dirname) && ((dirname[strlen(dirname)-1] == '\n') ||
+           (dirname[strlen(dirname)-1] == '\r')))
+      dirname[strlen(dirname)-1]='\0';
+    log_event(CPDEBUG, "output directory name generated", dirname);
+    mode=(0666&~conf.userumask);
   }
   free(user);
-  if (prepareuser(passwd, dirname)) return -2;
+  if (prepareuser(passwd, dirname)) {
+    free(dirname);
+    if (logfp!=NULL)
+      fclose(logfp);
+    return -2;
+  }
   log_event(CPDEBUG, "user information prepared", NULL);
 
-  size=strlen(CPSPOOL)+22;
+  size=strlen(conf.spool)+22;
   spoolfile=calloc(size, sizeof(char));
   if (spoolfile == NULL) {
     fputs("CUPS-PDF: failed to allocate memory\n", stderr);
     free(dirname);
-    fclose(logfp);
+    if (logfp!=NULL)
+      fclose(logfp);
     return -3;
   }
-  snprintf(spoolfile, size, "%s/cups2pdf-%i", CPSPOOL, (int) getpid());
+  snprintf(spoolfile, size, "%s/cups2pdf-%i", conf.spool, (int) getpid());
   log_event(CPDEBUG, "spoolfile name created", spoolfile);
 
   if (argc == 6) {
-    if (preparespoolfile(stdin, spoolfile, title, atoi(argv[1]), passwd)) {
+    if (preparespoolfile(stdin, spoolfile, title, argv[3], atoi(argv[1]), passwd)) {
       free(dirname);
       free(spoolfile);
-      fclose(logfp);
+      if (logfp!=NULL)
+        fclose(logfp);
       return -4;
     }
     log_event(CPDEBUG, "input data read from stdin", NULL);
   }
   else {
-    if (preparespoolfile(fopen(argv[6], "r"), spoolfile, title, atoi(argv[1]), passwd)) {
+    if (preparespoolfile(fopen(argv[6], "r"), spoolfile, title, argv[3], atoi(argv[1]), passwd)) {
       free(dirname);
       free(spoolfile);
-      fclose(logfp);
+      if (logfp!=NULL)
+        fclose(logfp);
       return -4;
     }
     log_event(CPDEBUG, "input data read from file", argv[6]);
@@ -463,40 +620,44 @@ int main(int argc, char *argv[]) {
     fputs("CUPS-PDF: failed to allocate memory\n", stderr);
     free(dirname);
     free(spoolfile);
-    fclose(logfp);
+    if (logfp!=NULL)
+      fclose(logfp);
     return -5;
   }
   snprintf(outfile, size, "%s/%s.pdf", dirname, title);
   log_event(CPDEBUG, "output filename created", outfile);
 
-  size=strlen(CPGSCALL)+strlen(CPGHOSTSCRIPT)+strlen(CPPDFVER)+strlen(outfile)+strlen(spoolfile)+6;
+  size=strlen(conf.gscall)+strlen(conf.ghostscript)+strlen(conf.pdfver)+strlen(outfile)+strlen(spoolfile)+6;
   gscall=calloc(size, sizeof(char));
   if (gscall == NULL) {
     fputs("CUPS-PDF: failed to allocate memory\n", stderr);
     free(dirname);
     free(spoolfile);
     free(outfile);
-    fclose(logfp);
+    if (logfp!=NULL)
+      fclose(logfp);
     return -6;
   }
-  snprintf(gscall, size, CPGSCALL, CPGHOSTSCRIPT, CPPDFVER, outfile, spoolfile);
+  snprintf(gscall, size, conf.gscall, conf.ghostscript, conf.pdfver, outfile, spoolfile);
   log_event(CPDEBUG, "ghostscript commandline built", gscall);
 
   unlink(outfile);
   log_event(CPDEBUG, "output file unlinked", outfile);
 
-  if (setenv("TMPDIR", CPGSTMP, 1)) {
-    log_event(CPERROR, "insufficient space in environment to set TMPDIR", CPGSTMP);
+  if (setenv("TMPDIR", conf.gstmp, 1)) {
+    log_event(CPERROR, "insufficient space in environment to set TMPDIR", conf.gstmp);
     free(dirname);
     free(spoolfile);
     free(outfile);
     free(gscall);
-    fclose(logfp);
+    if (logfp!=NULL)
+      fclose(logfp);
     return -7;
   }
   log_event(CPDEBUG, "TMPDIR set for GhostScript", getenv("TMPDIR"));
 
-  fflush(logfp);
+  if (logfp!=NULL)
+    fflush(logfp);
   pid=fork();
 
   if (!pid) {
@@ -511,10 +672,33 @@ int main(int argc, char *argv[]) {
     else 
       log_event(CPDEBUG, "UID set for current user", passwd->pw_name);
      
-    umask(0777);
+    umask(0077);
     size=system(gscall);
     snprintf(title,BUFSIZE,"%d",size);
     log_event(CPDEBUG, "ghostscript has finished", title);
+    if (chmod(outfile, mode)) {
+      log_event(CPERROR, "failed to set file mode for PDF file (non fatal)", outfile);
+      return -101;
+    }
+    log_event(CPDEBUG, "file mode set for user output", outfile);
+    
+    if (strlen(conf.postprocessing)) {
+      size=strlen(conf.postprocessing)+strlen(outfile)+strlen(passwd->pw_name)+3;
+      ppcall=calloc(size, sizeof(char));
+      if (ppcall == NULL) 
+        log_event(CPERROR, "failed to allocate memory for postprocessing (non fatal)", NULL);
+      else {
+        snprintf(ppcall, size, "%s %s %s", conf.postprocessing, outfile, passwd->pw_name);
+        log_event(CPDEBUG, "postprocessing commandline built", ppcall);
+        size=system(ppcall);
+        snprintf(title,BUFSIZE,"%d",size);
+        log_event(CPDEBUG, "postprocessing has finished", title);
+        free(ppcall);
+      }
+    }
+    else
+     log_event(CPDEBUG, "no postprocessing", NULL);
+
     return 0;
   }
   log_event(CPDEBUG, "waiting for child to exit", NULL);
@@ -525,17 +709,6 @@ int main(int argc, char *argv[]) {
   else 
     log_event(CPDEBUG, "spoolfile unlinked", spoolfile);
 
-  if (chmod(outfile, mode)) {
-    log_event(CPERROR, "failed to set file mode for PDF file", outfile);
-    free(dirname);
-    free(spoolfile);
-    free(outfile);
-    free(gscall);
-    fclose(logfp);
-    return -9;
-  }
-  log_event(CPDEBUG, "file mode set for user output", outfile);
- 
   free(dirname);
   free(spoolfile);
   free(outfile);
@@ -545,6 +718,7 @@ int main(int argc, char *argv[]) {
 
   log_event(CPSTATUS, "PDF creation successfully finished", passwd->pw_name);
 
-  fclose(logfp);
+  if (logfp!=NULL)
+    fclose(logfp);
   return 0;
 } 
